import contextvars
from typing import Any, Awaitable, Callable, Dict, Optional, Type

from arkitekt.api.schema import WidgetInput
from .errors import (
    StructureDefinitionError,
    StructureOverwriteError,
    StructureRegistryError,
)
from pydantic import BaseModel


current_structure_registry = contextvars.ContextVar(
    "current_structure_registry", default=None
)


async def id_shrink(self):
    return self.id


class StructureRegistry(BaseModel):
    copy_from_default: bool = False
    allow_overwrites: bool = True
    allow_auto_register: bool = False

    _identifier_expander_map: Dict[str, Callable[[str], Awaitable[Any]]] = {}
    _identifier_shrinker_map: Dict[str, Callable[[Any], Awaitable[str]]] = {}
    _structure_identifier_map: Dict[Type, str] = {}
    _structure_default_widget_map: Dict[Type, WidgetInput] = {}
    _identifier_structure_map: Dict[str, Type] = {}

    _token: contextvars.Token = None

    def get_expander_for_identifier(self, key):
        try:
            return self._identifier_expander_map[key]
        except KeyError as e:
            raise StructureRegistryError(f"{key} is not registered") from e

    def get_shrinker_for_identifier(self, key):
        try:
            return self._identifier_shrinker_map[key]
        except KeyError as e:
            raise StructureRegistryError(f"{key} is not registered") from e

    def register_expander(self, key, expander):
        self._identifier_expander_map[key] = expander

    def get_widget_input(self, cls) -> Optional[WidgetInput]:
        return self._structure_default_widget_map.get(cls, None)

    def get_identifier_for_structure(self, cls):
        try:
            return self._structure_identifier_map[cls]
        except KeyError as e:
            if self.allow_auto_register:
                try:
                    self.register_as_structure(cls)
                    return self._structure_identifier_map[cls]
                except StructureDefinitionError as e:
                    raise StructureDefinitionError(
                        f"{cls} was not registered and could not be registered automatically"
                    ) from e
            else:
                raise StructureRegistryError(
                    f"{cls} is not registered and allow_auto_register is set to False. Please make sure to register this type beforehand or set allow_auto_register to True"
                ) from e

    def register_as_structure(
        self,
        cls,
        identifier=None,
        expand=None,
        shrink=None,
        default_widget=None,
        id_shrink=True,
    ):
        if expand is None:
            if not hasattr(cls, "expand"):
                raise StructureDefinitionError(
                    f"You need to pass 'expand' method or {cls} needs to implement a expand method"
                )
            expand = cls.expand

        if shrink is None:
            if not hasattr(cls, "shrink"):
                if issubclass(cls, BaseModel):
                    if "id" in cls.__fields__:
                        cls.shrink = id_shrink
                    else:
                        raise StructureDefinitionError(
                            f"You need to pass 'shrink' method or {cls} needs to implement a shrink method. A BaseModel can be automatically shrinked by providing an id field"
                        )
                else:
                    raise StructureDefinitionError(
                        f"You need to pass 'shrink' method or {cls} needs to implement a shrink method"
                    )

            shrink = cls.shrink

        if identifier is None:
            if not hasattr(cls, "get_identifier"):
                raise StructureDefinitionError(
                    f"You need to pass 'identifier' or  {cls} needs to implement a get_identifier method"
                )
            identifier = cls.get_identifier()

        if default_widget:
            pass

        if identifier in self._identifier_structure_map and not self.allow_overwrites:
            raise StructureOverwriteError(
                f"{identifier} is already registered. Previously registered {self._identifier_structure_map[identifier]}"
            )

        self._identifier_expander_map[identifier] = expand
        self._identifier_shrinker_map[identifier] = shrink
        self._identifier_structure_map[identifier] = cls
        self._structure_identifier_map[cls] = identifier
        self._structure_default_widget_map[cls] = default_widget

    async def __aenter__(self):
        current_structure_registry.set(self)
        return self

    async def __exit__(self, exc_type, exc_val, exc_tb):
        current_structure_registry.set(None)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


DEFAULT_STRUCTURE_REGISTRY = None


def get_default_structure_registry():
    global DEFAULT_STRUCTURE_REGISTRY
    if not DEFAULT_STRUCTURE_REGISTRY:
        DEFAULT_STRUCTURE_REGISTRY = StructureRegistry()
    return DEFAULT_STRUCTURE_REGISTRY


def get_current_structure_registry(allow_default=True):
    return current_structure_registry.get(get_default_structure_registry())


def register_structure(
    identifier=None,
    expand=None,
    shrink=None,
    default_widget=None,
    registry: StructureRegistry = None,
):
    """A Decorator for registering a structure

    Args:
        identifier ([type], optional): [description]. Defaults to None.
        expand ([type], optional): [description]. Defaults to None.
        shrink ([type], optional): [description]. Defaults to None.
        default_widget ([type], optional): [description]. Defaults to None.
    """

    registry = registry or get_current_structure_registry()

    def func(cls):

        registry.register_as_structure(cls, identifier, expand, shrink, default_widget)
        return cls

    return func



from abc import abstractmethod
from typing import Any, List, Optional, Union
from arkitekt.messages import Assignation, Unassignation, Provision, Unprovision
from arkitekt.api.schema import ProvisionMode, ProvisionStatus, AssignationStatus
from koil.decorators import koilable


@koilable()
class AgentTransport:
    """Agent Transport

    A Transport is a means of communicating with an Agent. It is responsible for sending
    and receiving messages from the backend. It needs to implement the following methods:

    list_provision: Getting the list of active provisions from the backend. (depends on the backend)
    list_assignation: Getting the list of active assignations from the backend. (depends on the backend)

    change_assignation: Changing the status of an assignation. (depends on the backend)
    change_provision: Changing the status of an provision. (depends on the backend)

    broadcast: Configuring the callbacks for the transport on new assignation, unassignation provision and unprovison.

    if it is a stateful connection it can also implement the following methods:

    aconnect
    adisconnect

    """

    @property
    def connected(self):
        return NotImplementedError("Implement this method")

    @abstractmethod
    def abroadcast(
        self, message: Union[Assignation, Unassignation, Provision, Unprovision]
    ):
        raise NotImplementedError("This is an abstract Base Class")

    @abstractmethod
    async def list_provisions(
        self, exclude: Optional[ProvisionStatus] = None
    ) -> List[Provision]:
        raise NotImplementedError("This is an abstract Base Class")

    @abstractmethod
    async def change_provision(
        self,
        id: str,
        status: ProvisionStatus = None,
        message: str = None,
        mode: ProvisionMode = None,
    ):
        raise NotImplementedError("This is an abstract Base Class")

    @abstractmethod
    async def change_assignation(
        self,
        id: str,
        status: AssignationStatus = None,
        message: str = None,
        returns: List[Any] = None,
    ):
        raise NotImplementedError("This is an abstract Base Class")

    @abstractmethod
    async def list_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        raise NotImplementedError("This is an abstract Base Class")

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



from importlib import import_module
import logging
import sys
import os
from rich.console import Console
from arkitekt.actors.registry import get_current_actor_registry

logger = logging.getLogger(__name__)


class Run:
    def __init__(
        self,
        path=None,
        entrypoint="run",
    ) -> None:

        if path == ".":
            self.watch_path = os.getcwd()
            sys.path.insert(0, self.watch_path)
        else:
            self.watch_path = os.path.join(os.getcwd(), path)

        self.module_path = f"{path}.{entrypoint}" if path != "." else f"{entrypoint}"
        self.module = import_module(self.module_path)
        self.console = Console()
        os.environ["ARKITEKT_AGENT_DEBUG"] = "False"

    async def run(self):

        registry = get_current_actor_registry()

        if registry.has_actors():
            self.console.print(
                "We found functions that ought to be provide. Starting an Agent"
            )
            agent = ScriptAgent(registry=registry)
            await agent.aprovide()

        else:
            self.console.print(
                f"[red] No Functions in {self.module_path} that we can provide found. Exiting!"
            )


async def import_directory_and_start(path="watch", entrypoint="run"):
    host = Run(path=path)
    await host.run()



from arkitekt.postmans.stateful import StatefulPostman
from arkitekt.postmans.transport.websocket import WebsocketPostmanTransport
from fakts.config.base import Config
from fakts.fakts import Fakts
from herre.herre import Herre, current_herre


class PostmanConfig(Config):
    endpoint_url: str
    instance_id: str

    class Config:
        group = "arkitekt.postman"


class FaktsPostman(StatefulPostman):
    def __init__(self, fakts: Fakts = None) -> None:
        super().__init__(None)
        self.fakts = fakts

    def configure(
        self,
        config: PostmanConfig,
        herre: Herre,
    ) -> None:
        self.transport = WebsocketPostmanTransport(
            ws_url=config.endpoint_url,
            instance_id=config.instance_id,
            token_loader=herre.aget_token,
        )

    async def __aenter__(self):

        herre = current_herre.get()

        config = await PostmanConfig.from_fakts(fakts=self.fakts)
        self.configure(config, herre)

        return await super().__aenter__()



from arkitekt.agents.base import BaseAgent
from arkitekt.api.schema import AssignationStatus, ProvisionStatus
from arkitekt.messages import Assignation, Provision, Unassignation, Unprovision
from typing import Union
import asyncio
import logging

logger = logging.getLogger(__name__)


class StatefulAgent(BaseAgent):
    """An agent that tries to recover and
    take care of all the assignations and provisions

    Args:
        BaseAgent (_type_): _description_
    """

    async def process(
        self, message: Union[Assignation, Provision, Unassignation, Unprovision]
    ):
        logger.info(f"Agent received {message}")

        if isinstance(message, Assignation) or isinstance(message, Unassignation):
            if message.provision in self._provisionActorMap:
                actor = self._provisionActorMap[message.provision]
                await actor.apass(message)
            else:
                await self.transport.change_assignation(
                    message.assignation,
                    status=AssignationStatus.CRITICAL,
                    message="Actor that handles this provision is not available",
                )

        elif isinstance(message, Provision):
            if message.template in self._templateActorBuilderMap:
                actorBuilder = self._templateActorBuilderMap[message.template]
                self._provisionActorMap[message.provision] = actorBuilder(message, self)
                await self._provisionActorMap[message.provision].arun()
                logger.info("Actor started")
            else:
                logger.info("Actor not found")
                await self.transport.change_provision(
                    message.provision,
                    status=ProvisionStatus.DENIED,
                    message="No actor found on the provisioning Agent, this is most likely due to a change in this agent's configuration",
                )

        elif isinstance(message, Unprovision):
            await self._provisionActorMap[message.provision].astop()

        else:
            raise Exception(f"Unknown message type {type(message)}")

    async def __aexit__(self, exc_type, exc_val, exc_tb):

        cancelations = [actor.astop() for actor in self._provisionActorMap.values()]

        for c in cancelations:
            try:
                await c
            except asyncio.CancelledError:
                print(f"Cancelled Actor {c}")

        await super().__aexit__(exc_type, exc_val, exc_tb)



from arkitekt.agents.transport.base import AgentTransport
from arkitekt.agents.transport.protocols.agent_json import *
from arkitekt.messages import Assignation, Provision, Provision, Unprovision
from arkitekt.api.schema import AssignationStatus, ProvisionStatus
from typing import Any, List, Optional, Union
import asyncio
from koil import unkoil


class MockAgentTransport(AgentTransport):
    """A mock transport for an agent

    Args:
        AgentTransport (_type_): _description_
    """

    def __init__(self, *args, **kwargs):
        pass

    async def list_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        return []

    async def list_provisions(
        self, exclude: Optional[ProvisionStatus] = None
    ) -> List[Provision]:
        return []

    async def __aenter__(self):
        self._inqueue = asyncio.Queue()

    async def change_assignation(
        self,
        id: str,
        status: AssignationStatus = None,
        message: str = None,
        returns: List[Any] = None,
    ):
        await self._inqueue.put(
            AssignationChangedMessage(
                assignation=id, status=status, message=message, returns=returns
            )
        )

    async def change_provision(
        self,
        id: str,
        status: ProvisionStatus = None,
        message: str = None,
        mode: ProvisionMode = None,
    ):
        print(f"change_provision {id} {status} {message} {mode}")
        await self._inqueue.put(
            ProvisionChangedMessage(
                provision=id, status=status, message=message, mode=mode
            )
        )

    async def delay(
        self, message: Union[Assignation, Provision, Unprovision, Unassignation]
    ):
        await self.abroadcast(message)

    def sync_delay(
        self, message: Union[Assignation, Provision, Unprovision, Unassignation]
    ):
        return unkoil(self.delay, message)

    def sync_receive(self, *args, **kwargs):
        return unkoil(self.receive, *args, **kwargs)

    async def receive(self, timeout=None):
        if timeout:
            return await asyncio.wait_for(self._inqueue.get(), timeout)
        return await self._inqueue.get()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        for item in range(self._inqueue.qsize()):
            print(f"Flushing Item {item}")
            self._inqueue.task_done()



from arkitekt.rath import ArkitektRath, current_arkitekt_rath
from koil import unkoil
from koil.helpers import unkoil_gen


def execute(operation, variables, rath: ArkitektRath = None):
    return unkoil(aexecute, operation, variables, rath)


async def aexecute(operation, variables, rath: ArkitektRath = None):
    rath = rath or current_arkitekt_rath.get()
    x = await rath.aexecute(operation.Meta.document, variables)
    return operation(**x.data)


def subscribe(operation, variables, rath: ArkitektRath = None):
    return unkoil_gen(asubscribe, operation, variables, rath)


async def asubscribe(operation, variables, rath: ArkitektRath = None):
    rath = rath or current_arkitekt_rath.get()
    async for event in rath.asubscribe(operation.Meta.document, variables):
        yield operation(**event.data)



class ExpandingError(Exception):
    pass


class ShrinkingError(Exception):
    pass


class StructureRegistryError(Exception):
    pass


class StructureOverwriteError(StructureRegistryError):
    pass


class StructureDefinitionError(StructureRegistryError):
    pass



from typing import Dict, Union
from arkitekt.agents.transport.base import AgentTransport
from arkitekt.structures.registry import StructureRegistry
from arkitekt.rath import ArkitektRath
import asyncio
import logging
from arkitekt.api.schema import (
    AssignationStatus,
    ProvisionMode,
    ProvisionStatus,
    aget_template,
    TemplateFragment,
)
from arkitekt.messages import Assignation, Provision, Unassignation
from arkitekt.actors.errors import UnknownMessageError
from arkitekt.actors.vars import current_provision_context

logger = logging.getLogger(__name__)


class Agent:
    transport: AgentTransport
    rath: ArkitektRath


class Actor:
    template: TemplateFragment
    transport: AgentTransport
    rath: ArkitektRath

    def __init__(
        self,
        provision: Provision,
        agent: Agent,
        *args,
        strict=False,
        expand_inputs=True,
        shrink_outputs=True,
        structure_registry: StructureRegistry = None,
        debug=False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict
        self.expand_inputs = expand_inputs
        self.shrink_outputs = shrink_outputs
        self.structure_registry = structure_registry
        self.runningAssignments: Dict[
            str, asyncio.Task
        ] = {}  # Running assignments indexed by assignment reference
        self.debug = debug
        self.provision = provision
        self.agent = agent
        self.transport = agent.transport
        self.rath = agent.rath

    async def on_provide(self, provision: Provision, template: TemplateFragment):
        return None

    async def on_unprovide(self):
        return None

    async def on_assign(self, assignation: Assignation):
        raise (
            "Needs to be owerwritten in Actor Subclass. Never use this class directly"
        )

    async def apass(self, message: Union[Assignation, Unassignation]):
        await self.in_queue.put(message)

    async def arun(self):
        self.in_queue = asyncio.Queue()
        self.template = await aget_template(id=self.provision.template, rath=self.rath)
        self.provision_task = asyncio.create_task(self.alisten())

    async def astop(self):
        self.provision_task.cancel()

        try:
            await self.provision_task
        except asyncio.CancelledError:
            print("Provision was cancelled")

    async def alisten(self):
        try:
            self.template = await aget_template(
                id=self.provision.template, rath=self.rath
            )

            await self.transport.change_provision(
                self.provision.provision, status=ProvisionStatus.PROVIDING
            )

            prov_context = await self.on_provide(self.provision, self.template)
            current_provision_context.set(prov_context)

            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.ACTIVE,
                mode=ProvisionMode.DEBUG if self.debug else ProvisionMode.PRODUCTION,
            )
            logger.info(f"Actor for {self.provision}: Is now active")

            while True:
                message = await self.in_queue.get()
                logger.info(f"Actor for {self.provision}: Received {message}")

                if isinstance(message, Assignation):
                    task = asyncio.create_task(self.on_assign(message))
                    self.runningAssignments[message.assignation] = task

                elif isinstance(message, Unassignation):
                    if message.assignation in self.runningAssignments:
                        task = self.runningAssignments[message.assignation]
                        if not task.done():
                            print("Cancelling task")
                            task.cancel()
                        else:
                            logger.error("Task was already done")
                    else:
                        await self.transport.change_assignation(
                            status=AssignationStatus.CRITICAL,
                            message="Task was never assigned",
                        )
                else:
                    raise UnknownMessageError(f"{message}")

        except Exception as e:
            logger.exception("Actor failed", exc_info=True)
            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CRITICAL,
                message=repr(e),
            )

            current_provision_context.set(None)

        except asyncio.CancelledError as e:
            print("We are getting cancelled here?")
            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CANCELING,
                message=repr(e),
            )

            await self.on_unprovide()

            logger.info("Doing Whatever needs to be done to cancel!")
            await self.transport.change_provision(
                self.provision.provision,
                status=ProvisionStatus.CANCELLED,
                message=str(e),
            )

            cancel_assignations = [i.cancel() for i in self.runningAssignments.values()]

            for i in self.runningAssignments.values():
                try:
                    await i
                except asyncio.CancelledError:
                    pass

            current_provision_context.set(None)
            raise e

    async def __aenter__(self):
        await self.arun()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.astop()



from enum import Enum
from typing import Callable
import inflection
from arkitekt.api.schema import (
    ArgPortInput,
    DefinitionInput,
    KwargPortInput,
    NodeTypeInput,
    ReturnPortInput,
)
import inspect
from docstring_parser import parse
from arkitekt.definition.errors import DefinitionError

from arkitekt.structures.registry import (
    StructureRegistry,
)


def convert_arg_to_argport(
    cls, registry: StructureRegistry, widget=None, key=None
) -> ArgPortInput:
    """
    Convert a class to an ArgPort
    """

    widget = widget or registry.get_widget_input(cls)

    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_arg_to_argport(cls.__args__[0], registry)
            return ArgPortInput(
                typename="ListArgPort", widget=widget, key=key, child=child
            )

        if cls._name == "Dict":
            child = convert_arg_to_argport(cls.__args__[1], registry)
            return ArgPortInput(
                typename="DictArgPort", widget=widget, key=key, child=child
            )

    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool):
            return ArgPortInput(
                typename="BoolArgPort", widget=widget, key=key
            )  # catch bool is subclass of int
        if issubclass(cls, Enum):
            return ArgPortInput(
                typename="EnumArgPort",
                widget=widget,
                key=key,
                options={key: value._value_ for key, value in cls.__members__.items()},
            )
        if issubclass(cls, int):
            return ArgPortInput(typename="IntArgPort", widget=widget, key=key)
        if issubclass(cls, str):
            return ArgPortInput(typename="StringArgPort", widget=widget, key=key)

    identifier = registry.get_identifier_for_structure(cls)

    return ArgPortInput(
        typename="StructureArgPort",
        identifier=identifier,
        widget=widget,
        key=key,
    )


def convert_kwarg_to_kwargport(
    cls, registry: StructureRegistry, widget=None, key=None, default=None
) -> ArgPortInput:
    """
    Convert a class to an ArgPort
    """
    widget = widget or registry.get_widget_input(cls)

    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_kwarg_to_kwargport(cls.__args__[0], registry)
            return KwargPortInput(
                typename="ListKwargPort",
                widget=widget,
                key=key,
                child=child,
                defaultList=default,
            )

        if cls._name == "Dict":
            child = convert_kwarg_to_kwargport(cls.__args__[1], registry)
            return KwargPortInput(
                typename="DictKwargPort",
                widget=widget,
                key=key,
                child=child,
                defaultDict=default,
            )

    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool) or isinstance(default, bool):
            print(f"RUnning into this {default}")
            t = KwargPortInput(
                typename="BoolKwargPort", widget=widget, key=key, defaultBool=default
            )  # catch bool is subclass of int
            print(t.dict(by_alias=True))
            return t

        if issubclass(cls, Enum) or isinstance(default, Enum):
            return KwargPortInput(
                typename="EnumKwargPort",
                widget=widget,
                key=key,
                options={key: value._value_ for key, value in cls.__members__.items()},
                defaultEnum=default,
            )
        if issubclass(cls, int) or isinstance(default, int):
            return KwargPortInput(
                typename="IntKwargPort", widget=widget, key=key, defaultInt=default
            )
        if issubclass(cls, str) or isinstance(default, str):
            return KwargPortInput(
                typename="StringKwargPort",
                widget=widget,
                key=key,
                defaultString=default,
            )

    identifier = registry.get_identifier_for_structure(cls)

    return KwargPortInput(
        typename="StructureKwargPort",
        identifier=identifier,
        widget=widget,
        key=key,
        defaultID=default,
    )


def convert_return_to_returnport(
    cls, registry: StructureRegistry, key=None
) -> ReturnPortInput:
    """
    Convert a class to an ArgPort
    """

    if hasattr(cls, "_name"):
        # We are dealing with a Typing Var?
        if cls._name == "List":
            child = convert_return_to_returnport(cls.__args__[0], registry)
            return ReturnPortInput(
                typename="ListReturnPort",
                key=key,
                child=child,
            )

        if cls._name == "Dict":
            child = convert_return_to_returnport(cls.__args__[1], registry)
            return ReturnPortInput(
                typename="DictReturnPort",
                key=key,
                child=child,
            )

    if inspect.isclass(cls):
        # Generic Cases

        if issubclass(cls, bool):
            return ReturnPortInput(
                typename="BoolReturnPort", key=key
            )  # catch bool is subclass of int
        if issubclass(cls, Enum):
            return ReturnPortInput(
                typename="EnumReturnPort",
                key=key,
                options={key: value._value_ for key, value in cls.__members__.items()},
            )
        if issubclass(cls, int):
            return ReturnPortInput(
                typename="IntReturnPort",
                key=key,
            )
        if issubclass(cls, str):
            return ReturnPortInput(
                typename="StringReturnPort",
                key=key,
            )

    identifier = registry.get_identifier_for_structure(cls)

    return ReturnPortInput(
        typename="StructureReturnPort",
        identifier=identifier,
        key=key,
    )


def prepare_definition(
    function: Callable,
    widgets={},
    allow_empty_doc=False,
    interfaces=[],
    structure_registry: StructureRegistry = None,
) -> DefinitionInput:
    """Define

    Define a functions in the context of arnheim and
    return it as a Node. Attention this node is not yet
    hosted on Arkitekt (doesn't have an id). So make sure
    to save this node before calling it anywhere

    Args:
        function (): The function you want to define
    """

    assert structure_registry is not None, "You need to pass a StructureRegistry"

    is_generator = inspect.isasyncgenfunction(function) or inspect.isgeneratorfunction(
        function
    )

    sig = inspect.signature(function)

    # Generate Args and Kwargs from the Annotation
    args = []
    kwargs = []
    returns = []

    function_ins_annotation = sig.parameters
    for key, value in function_ins_annotation.items():

        widget = widgets.get(key, None)
        cls = value.annotation

        try:
            if value.default == inspect.Parameter.empty:
                # This Parameter is an Argument
                args.append(
                    convert_arg_to_argport(
                        cls, structure_registry, widget=widget, key=key
                    )
                )
            else:
                kwargs.append(
                    convert_kwarg_to_kwargport(
                        cls,
                        structure_registry,
                        widget=widget,
                        key=key,
                        default=value.default,
                    )
                )
        except Exception as e:
            raise DefinitionError(
                f"Could not convert Argument of function {function.__name__} to ArgPort: {value}"
            ) from e

    function_outs_annotation = sig.return_annotation

    if hasattr(function_outs_annotation, "_name"):

        if function_outs_annotation._name == "Tuple":
            try:
                for cls in function_outs_annotation.__args__:
                    returns.append(
                        convert_return_to_returnport(cls, structure_registry)
                    )
            except Exception as e:
                raise DefinitionError(
                    f"Could not convert Return of function {function.__name__} to ArgPort: {cls}"
                ) from e
        else:
            try:
                returns.append(
                    convert_return_to_returnport(
                        function_outs_annotation, structure_registry
                    )
                )  # Other types will be converted to normal lists and shit
            except Exception as e:
                raise DefinitionError(
                    f"Could not convert Return of function {function.__name__} to ArgPort: {function_outs_annotation}"
                ) from e
    else:
        # We are dealing with a non tuple return
        if function_outs_annotation is None:
            pass

        elif function_outs_annotation.__name__ != "_empty":  # Is it not empty
            returns.append(
                convert_return_to_returnport(
                    function_outs_annotation, structure_registry
                )
            )

    # Documentation Parsing

    # Docstring Parser to help with descriptions
    docstring = parse(function.__doc__)
    if docstring.long_description is None:
        assert (
            allow_empty_doc is not False
        ), f"We don't allow empty documentation for function {function.__name__}. Please Provide"

    name = docstring.short_description or function.__name__
    interface = inflection.underscore(function.__name__)  # convert this to camelcase
    description = docstring.long_description or "No Description"

    doc_param_map = {
        param.arg_name: {
            "description": param.description
            if not param.description.startswith("[")
            else None,
        }
        for param in docstring.params
    }

    if docstring.returns:
        return_description = docstring.returns.description
        seperated_list = return_description.split(",")
        assert len(returns) == len(
            seperated_list
        ), f"Length of Description and Returns not Equal: If you provide a Return Annotation make sure you seperate the description for each port with ',' Return Description {return_description} Returns: {returns}"
        for index, doc in enumerate(seperated_list):
            returns[index].description = doc

    # TODO: Update with documentatoin.... (Set description for portexample)
    for port in args:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description

    for port in kwargs:
        if port.key in doc_param_map:
            updates = doc_param_map[port.key]
            port.description = updates["description"] or port.description

    x = DefinitionInput(
        **{
            "name": name,
            "interface": interface,
            "package": "test",  # TODO: IMplement correctly
            "description": description,
            "args": args,
            "kwargs": kwargs,
            "returns": returns,
            "type": NodeTypeInput.GENERATOR if is_generator else NodeTypeInput.FUNCTION,
            "interfaces": interfaces,
        }
    )

    return x



import ast
from typing import List, Optional

from black import os
from turms.config import GeneratorConfig
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.language.ast import OperationDefinitionNode, OperationType
from turms.globals import FRAGMENT_CLASS_MAP, FRAGMENT_DOCUMENT_MAP
from turms.parser.recurse import recurse_annotation
from turms.plugins.base import Plugin
from pydantic import BaseModel
from graphql.error.graphql_error import GraphQLError
from graphql.language.ast import (
    FieldNode,
    FragmentDefinitionNode,
    OperationDefinitionNode,
    OperationType,
)
from graphql.utilities.build_client_schema import GraphQLSchema
from graphql.utilities.get_operation_root_type import get_operation_root_type
from graphql.utilities.type_info import get_field_def

from pydantic.main import BaseModel
from graphql import language, parse, validate
from turms.plugins.funcs import (
    OperationsFuncPluginConfig,
    generate_operation_func,
)
from turms.plugins.operation import OperationsPluginConfig, generate_query
from turms.utils import (
    generate_typename_field,
    fragment_searcher,
    replace_iteratively,
)
import logging
import glob

logger = logging.getLogger(__name__)


class StructurePluginsConfig(BaseModel):
    query_bases: List[str] = ["turms.types.herre.GraphQLQuery"]
    mutation_bases: List[str] = ["turms.types.herre.GraphQLMutation"]
    structure_bases: List[str] = ["arkitekt.types.GraphQLStructure"]

    fragment_bases: List[str] = ["turms.types.object.GraphQLObject"]
    structure_bases: List[str] = [
        "turms.types.object.GraphQLObject",
        "arkitekt.types.GraphQLStructure",
    ]

    prepend: str = ""
    append: str = ""
    structure_glob: Optional[str]
    out_dir: str = "api/structures"


def generate_fragment_tree(
    f: FragmentDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: StructurePluginsConfig,
    structure_name: str,
):
    tree = []
    fields = []
    type = client_schema.get_type(f.type_condition.name.value)

    name = f"{config.prepend_fragment}{f.name.value}{config.append_fragment}"

    fields += [generate_typename_field(type.name)]

    for field in f.selection_set.selections:

        field_definition = get_field_def(client_schema, type, field)
        assert field_definition, "Couldn't find field definition"

        target = (
            field.alias.value
            if hasattr(field, "alias") and field.alias
            else field.name.value
        )

        fields.append(
            ast.AnnAssign(
                target=ast.Name(target, ctx=ast.Store()),
                annotation=recurse_annotation(
                    field,
                    field_definition.type,
                    client_schema,
                    config,
                    tree,
                    parent_name=target.capitalize(),
                ),
                simple=1,
            )
        )

    FRAGMENT_DOCUMENT_MAP[f.name.value] = language.print_ast(f)
    FRAGMENT_CLASS_MAP[f.name.value] = name

    if f.name.value == structure_name:

        print(f"Generating Structure {structure_name}")
        tree.append(
            ast.ClassDef(
                name,
                bases=[
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.structure_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields,
            )
        )

    else:
        tree.append(
            ast.ClassDef(
                name,
                bases=[
                    ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                    for base in plugin_config.fragment_bases
                ],
                decorator_list=[],
                keywords=[],
                body=fields,
            )
        )
    return tree


def generate_structure(
    o: OperationDefinitionNode,
    client_schema: GraphQLSchema,
    config: GeneratorConfig,
    plugin_config: StructurePluginsConfig,
):

    tree = []

    x = get_operation_root_type(client_schema, o)
    query_fields = []
    name = f"{plugin_config.prepend}{o.name.value}{plugin_config.append}"

    for field_node in o.selection_set.selections:

        field_node: FieldNode = field_node
        field_definition = get_field_def(client_schema, x, field_node)
        assert field_definition, "Couldn't find field definition"

        target = (
            field_node.alias.value
            if hasattr(field_node, "alias") and field_node.alias
            else field_node.name.value
        )

        query_fields += [
            ast.AnnAssign(
                target=ast.Name(target, ctx=ast.Store()),
                annotation=recurse_annotation(
                    field_node,
                    field_definition.type,
                    client_schema,
                    config,
                    tree,
                    parent_name=name,
                ),
                simple=1,
            )
        ]

    query_document = language.print_ast(o)
    z = fragment_searcher.findall(query_document)

    merged_document = replace_iteratively(query_document)

    query_fields += [
        ast.ClassDef(
            "Meta",
            bases=[],
            decorator_list=[],
            keywords=[],
            body=[
                ast.Assign(
                    targets=[ast.Name(id="ward", ctx=ast.Store())],
                    value=ast.Constant(value=str("arkitekt")),
                ),
                ast.Assign(
                    targets=[ast.Name(id="document", ctx=ast.Store())],
                    value=ast.Constant(value=merged_document),
                ),
            ],
        )
    ]
    tree.append(
        ast.ClassDef(
            name,
            bases=[
                ast.Name(id=base.split(".")[-1], ctx=ast.Load())
                for base in plugin_config.query_bases
            ],
            decorator_list=[],
            keywords=[],
            body=query_fields,
        )
    )

    return tree


class StructurePlugin(Plugin):
    def __init__(self, config=None, **data):
        self.plugin_config = config or StructurePluginsConfig(**data)

    def generate_imports(
        self, config: GeneratorConfig, client_schema: GraphQLSchema
    ) -> List[ast.AST]:
        imports = []

        all_bases = (
            self.plugin_config.query_bases
            + self.plugin_config.mutation_bases
            + self.plugin_config.structure_bases
        )

        for item in all_bases:
            imports.append(
                ast.ImportFrom(
                    module=".".join(item.split(".")[:-1]),
                    names=[ast.alias(name=item.split(".")[-1])],
                    level=0,
                )
            )

        return imports

    def generate_body(
        self, client_schema: GraphQLSchema, config: GeneratorConfig
    ) -> List[ast.AST]:

        files = glob.glob(self.plugin_config.structure_glob, recursive=True)

        errors: List[GraphQLError] = []

        func_config = OperationsFuncPluginConfig(generate_sync=False, prepend_async="")
        operations_plugin_config = OperationsPluginConfig()

        for file in files:
            with open(file, "r") as f:

                filename = os.path.basename(file).split(".")[0]

                print(f)
                structure_tree = []

                dsl = f.read()

                nodes = parse(dsl)
                errors = validate(client_schema, nodes)

                if len(errors) > 0:
                    raise Exception(errors)

                definitions = nodes.definitions

                fragments = [
                    node
                    for node in definitions
                    if isinstance(node, FragmentDefinitionNode)
                ]

                operations = [
                    node
                    for node in definitions
                    if isinstance(node, OperationDefinitionNode)
                ]

                # first lets find the fragment that is going to be the structure
                expand_ops = [op for op in operations if op.name.value == "__expand__"]
                assert (
                    len(expand_ops) == 1
                ), "There can only be one __expand__ operation"

                expand_op = expand_ops[0]

                structure_name = (
                    expand_op.selection_set.selections[0]
                    .selection_set.selections[0]
                    .name.value
                )

                for fragment in fragments:

                    structure_tree += generate_fragment_tree(
                        fragment,
                        client_schema,
                        config,
                        self.plugin_config,
                        structure_name,
                    )

                for operation in operations:
                    if operation.operation == OperationType.QUERY:
                        structure_tree += generate_query(
                            operation, client_schema, config, operations_plugin_config
                        )

                for operation in operations:
                    structure_tree += generate_operation_func(
                        operation, client_schema, config, func_config
                    )

                md = ast.Module(body=structure_tree, type_ignores=[])
                generated = ast.unparse(ast.fix_missing_locations(md))

                if not os.path.isdir(self.plugin_config.out_dir):
                    os.makedirs(self.plugin_config.out_dir)

                with open(
                    os.path.join(self.plugin_config.out_dir, f"{filename}.py"), "w"
                ) as f:
                    f.write(generated)

        return []



from typing import List
from arkitekt.api.schema import (
    ArgPortFragmentBase,
    BoolKwargPortFragment,
    KwargPortFragmentBase,
    ListArgPortFragment,
    ListKwargPortFragment,
    ListReturnPortFragment,
    ListReturnPortFragmentChildStructureReturnPortFragment,
    NodeFragment,
    ReturnPortFragmentBase,
    StringReturnPortFragment,
    StructureArgPortFragment,
    StructureReturnPortFragment,
    find,
)
from black import format_str, FileMode
import ast
import re


def clean(varStr):
    return re.sub("\W|^(?=\d)", "_", varStr)


identifier_class_map = {
    "representation": "mikro.Representation",
    "omerofile": "mikro.OmeroFile",
    "sample": "mikro.Sample",
    "imageplus": "mikro.ImagePlus",
    "metric": "mikro.Metric",
    "thumbnail": "mikro.Thumbnail",
}

shrinker_map = {
    "representation": "Representation",
    "omerofile": "OmeroFile",
    "sample": "Sample",
    "imageplus": "ImagePlus",
    "metric": "Metric",
    "thumbnail": "Thumbnail",
}

expander_map = shrinker_map


def generate_imports():

    imports = []

    for key, item in identifier_class_map.items():
        imports.append(
            ast.ImportFrom(
                module=".".join(item.split(".")[:-1]),
                names=[ast.alias(name=item.split(".")[-1])],
                level=0,
            )
        )

    imports += [
        ast.ImportFrom(
            module="enum",
            names=[ast.alias(name="Enum")],
            level=0,
        ),
        ast.ImportFrom(
            module="pydantic.fields",
            names=[ast.alias(name="Field")],
            level=0,
        ),
        ast.ImportFrom(
            module="typing",
            names=[
                ast.alias(name="Optional"),
                ast.alias(name="List"),
                ast.alias(name="Dict"),
                ast.alias(name="Union"),
                ast.alias(name="Literal"),
            ],
            level=0,
        ),
    ]

    return imports


def generate_node_args(
    args: List[ArgPortFragmentBase], kwargs: List[KwargPortFragmentBase]
):

    pos_args = []

    for arg in args:
        if isinstance(arg, StructureArgPortFragment):
            pos_args.append(
                ast.arg(
                    arg=arg.key or "fail",
                    annotation=ast.Name(
                        id=shrinker_map[arg.identifier],
                        ctx=ast.Load(),
                    ),
                )
            )

        if isinstance(arg, ListArgPortFragment):
            pos_args.append(
                ast.arg(
                    arg=arg.key or "fail",
                    annotation=ast.Subscript(
                        value=ast.Name("List", ctx=ast.Load()),
                        slice=ast.Name(
                            id=shrinker_map[arg.child.identifier],
                            ctx=ast.Load(),
                        ),
                    ),
                )
            )

    kw_args = []
    kw_values = []

    for kwarg in kwargs:
        if isinstance(kwarg, BoolKwargPortFragment):
            kw_args.append(
                ast.arg(
                    arg=kwarg.key or "fail",
                    annotation=ast.Name(
                        id="bool",
                        ctx=ast.Load(),
                    ),
                )
            )
            kw_values.append(ast.Constant(value=kwarg.defaultBool))

        if isinstance(kwarg, ListKwargPortFragment):
            kw_args.append(
                ast.arg(
                    arg=kwarg.key or "fail",
                    annotation=ast.Subscript(
                        value=ast.Name("List", ctx=ast.Load()),
                        slice=ast.Name(
                            id=shrinker_map[kwarg.child.identifier],
                            ctx=ast.Load(),
                        ),
                    ),
                )
            )
            kw_values.append(ast.Constant(value=None))

    return ast.arguments(
        args=pos_args + kw_args,
        posonlyargs=[],
        kwonlyargs=[],
        kw_defaults=[],
        defaults=kw_values,
    )


def generate_node_returns(returns: List[ReturnPortFragmentBase]):

    if len(returns) == 0:
        return ast.Name(id="None")

    if len(returns) == 1:
        re = returns[0]
        if isinstance(re, StringReturnPortFragment):
            return ast.Name(id="str", ctx=ast.Load())
        if isinstance(re, StructureReturnPortFragment):
            return ast.Name(id=expander_map[re.identifier], ctx=ast.Load())
        if isinstance(re, ListReturnPortFragment):

            if isinstance(
                re.child, ListReturnPortFragmentChildStructureReturnPortFragment
            ):
                return ast.Subscript(
                    value=ast.Name("List", ctx=ast.Load()),
                    slice=ast.Name(
                        id=shrinker_map[re.child.identifier],
                        ctx=ast.Load(),
                    ),
                )
        else:
            return ast.Name(id="ss", ctx=ast.Load())

    tuple_back = []

    return ast.Name(id="xxx", ctx=ast.Load())


def generate_node_doc(node: NodeFragment):

    description = f"{node.name}\n\n{node.description}\n"

    if len(node.args) > 0 or len(node.kwargs) > 0:
        description += "\nArguments:\n"
        for arg in node.args:
            description += f"\t{arg.key} ({arg.typename}) {arg.description}\n"
        for kwarg in node.kwargs:
            description += (
                f"\t{kwarg.key} ({kwarg.typename}, Optional) {kwarg.description}\n"
            )

    if len(node.returns) > 0:
        description += "\nReturns:\n"
        for re in node.returns:
            description += f"\t{re.key} ({re.typename}) {re.description}\n"

    return ast.Expr(value=ast.Constant(value=description))


def node_to_ast(node: NodeFragment):

    return [
        ast.AsyncFunctionDef(
            name=f"{clean(node.interface)}",
            args=generate_node_args(node.args, node.kwargs),
            body=[
                generate_node_doc(node),
                ast.Return(value=ast.Constant(value=None)),
            ],
            decorator_list=[],
            returns=generate_node_returns(node.returns),
        )
    ]


def generate_node_func(
    package: str = "", interface: str = "", file="generated_node.py"
):

    global_tree = []

    global_tree = generate_imports()

    node = find(package=package, interface=interface)

    global_tree += node_to_ast(node)

    md = ast.Module(body=global_tree, type_ignores=[])
    generated = ast.unparse(ast.fix_missing_locations(md))

    res = format_str(generated, mode=FileMode())

    with open(file, "w") as f:
        f.write(res)



from arkitekt.postmans.vars import current_postman
from arkitekt.structures.registry import get_current_structure_registry
from koil.decorators import koilable
from .stateful import StatefulPostman
from arkitekt.api.schema import AssignationStatus, ReservationStatus, ReserveParamsInput
import asyncio
from arkitekt.traits.node import Reserve
from koil import unkoil
import logging
from arkitekt.structures.serialization.postman import shrink_inputs, expand_outputs


logger = logging.getLogger(__name__)


@koilable(fieldname="_koil", add_connectors=True)
class ReservationContract:
    def __init__(
        self,
        node: Reserve,
        postman: StatefulPostman = None,
        params: ReserveParamsInput = None,
        auto_unreserve: bool = True,
    ):
        self.postman = postman
        if isinstance(node, Reserve):
            self.node = node
        else:
            raise NotImplementedError("You need to pass a Reservable Instance")
        self.params = params or ReserveParamsInput()
        self.reservation = None
        self.postman = None
        self.auto_unreserve = auto_unreserve

    async def aassign(self, *args, structure_registry=None, **kwargs):
        assert self.reservation, "We never entered the context manager"
        assert (
            self.reservation.status == ReservationStatus.ACTIVE
        ), "Reservation is not active"

        structure_registry = structure_registry or get_current_structure_registry()
        _ass_queue = asyncio.Queue()

        shrinked_args, shrinked_kwargs = await shrink_inputs(
            self.node, args, kwargs, structure_registry=structure_registry
        )

        ass = await self.postman.aassign(
            self.reservation.reservation, shrinked_args, shrinked_kwargs
        )
        self.postman.register_assignation_queue(ass.assignation, _ass_queue)
        logger.info(f"Listening to assignation updates for {ass}")
        while True:  # Waiting for assignation
            ass = await _ass_queue.get()
            logger.info(f"Reservation Context: {ass}")
            if ass.status == AssignationStatus.RETURNED:
                self.postman.unregister_assignation_queue(ass.assignation)
                return await expand_outputs(
                    self.node, ass.returns, structure_registry=structure_registry
                )

            if ass.status == AssignationStatus.CRITICAL:
                self.postman.unregister_assignation_queue(ass.assignation)
                raise Exception(f"Critical error: {ass.message}")

    def assign(self, *args, **kwargs):
        return unkoil(self.aassign, *args, **kwargs)

    async def watch_updates(self):
        try:
            if self.reservation.status == ReservationStatus.ACTIVE:
                self._enter_future.set_result(True)

            while True:
                self.reservation = await self._updates_queue.get()
                if self.reservation.status == ReservationStatus.ACTIVE:
                    if self._enter_future and not self._enter_future.done():
                        self._enter_future.set_result(True)

        except asyncio.CancelledError:
            pass

    async def __aenter__(self):
        self.postman = self.postman or current_postman.get()
        logger.info(f"Trying to reserve {self.node}")
        self.reservation = await self.postman.areserve(self.node.id, self.params)
        logger.info(f"Waiting for Reservation {self.reservation}")
        self._updates_queue = asyncio.Queue()
        self._enter_future = asyncio.Future()
        self.postman.register_reservation_queue(
            self.reservation.reservation, self._updates_queue
        )
        self.updates_watcher = asyncio.create_task(self.watch_updates())
        await self._enter_future  # Waiting to enter

        return self

    async def __aexit__(self, *args, **kwargs):
        self.updates_watcher.cancel()

        try:
            await self.updates_watcher
        except asyncio.CancelledError:
            pass

        if self.auto_unreserve:
            await asyncio.wait_for(
                self.postman.aunreserve(self.reservation.reservation), timeout=1
            )
        self.postman.unregister_reservation_queue(self.reservation.reservation)

    def __enter__(self) -> "ReservationContract":
        ...


def use(
    node: str,
    params: ReserveParamsInput = None,
    postman: StatefulPostman = None,
    auto_unreserve=True,
) -> ReservationContract:
    return ReservationContract(
        node, postman=postman, params=params, auto_unreserve=auto_unreserve
    )



from arkitekt.agents.errors import AgentException


class AgentTransportException(AgentException):
    """
    Base class for all exceptions raised by the Agent Transport.
    """



class ProvisionListDeniedError(AgentTransportException):
    """
    Raised when the backend is not able to list the provisions.
    """



class AssignationListDeniedError(AgentTransportException):
    """
    Raised when the backend is not able to list the assignations.
    """




from typing import Any, Dict, List, Optional, TypeVar
from pydantic import BaseModel
from arkitekt.api.schema import (
    ProvisionStatus,
    ReservationStatus,
    AssignationStatus,
)

T = TypeVar("T", bound=BaseModel)


class UpdatableModel(BaseModel):
    pass

    def update(self: T, other: BaseModel, in_place=True) -> Optional[T]:
        if in_place:
            for key, value in other.dict().items():
                if key in self.__fields__:
                    if value != None:  # None is not a valid update!
                        setattr(self, key, value)
        else:
            copy = self.copy()
            copy.update(other)
            return copy


class Assignation(UpdatableModel):
    assignation: str
    provision: Optional[str]
    reservation: Optional[str]
    args: Optional[List[Any]]
    kwargs: Optional[Dict[str, Any]]
    returns: Optional[List[Any]]
    persist: Optional[bool]
    log: Optional[bool]
    status: Optional[AssignationStatus]
    message: Optional[str]


class Unassignation(UpdatableModel):
    assignation: str
    provision: Optional[str]


class Provision(UpdatableModel):
    provision: str
    template: Optional[str]
    status: Optional[ProvisionStatus]


class Unprovision(UpdatableModel):
    provision: str


class Reservation(UpdatableModel):
    reservation: str
    template: Optional[str]
    node: Optional[str]
    status: Optional[ReservationStatus] = None


class Unreservation(BaseModel):
    reservation: str



import contextvars
import threading
from typing import Any, Coroutine, Dict, List, Optional
from arkitekt.actors.errors import ThreadedActorCancelled
from arkitekt.messages import Assignation
from arkitekt.api.schema import AssignationStatus
from arkitekt.actors.base import Actor
from concurrent.futures import ThreadPoolExecutor
import asyncio
from arkitekt.structures.serialization.actor import expand_inputs, shrink_outputs
from arkitekt.actors.vars import (
    current_assignation,
    current_janus_queue,
)
import logging
import janus
from koil.vars import current_cancel_event
from koil.helpers import run_spawned

logger = logging.getLogger(__name__)


class FunctionalActor(Actor):
    pass

    def __init__(
        self, *args, assign=None, on_provide=None, on_unprovide=None, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.assign = assign or self.assign
        self.on_provide = on_provide or self.on_provide
        self.on_unprovide = on_unprovide or self.on_unprovide

    def assign(self, *args, **kwargs):
        raise NotImplementedError("Please overwrite this")


class FunctionalFuncActor(FunctionalActor):
    assign: Coroutine[Any, Any, Any]

    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, assignation: Assignation):
        try:
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    assignation.args,
                    assignation.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (assignation.args, assignation.kwargs)
            )

            current_assignation.set(assignation)

            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            returns = await self.assign(*args, **kwargs)

            current_assignation.set(None)

            returns = (
                await shrink_outputs(
                    self.template.node,
                    returns,
                    structure_registry=self.structure_registry,
                )
                if self.shrink_outputs
                else returns
            )

            print("Should be returning", returns)

            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.RETURNED,
                returns=returns,
            )

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                assignation.assignation, status=AssignationStatus.CANCELLED
            )

        except Exception as e:
            logger.exception(e)
            await self.transport.change_assignation(
                assignation.assignation,
                status=AssignationStatus.CRITICAL,
                message=repr(e),
            )


class FunctionalGenActor(FunctionalActor):
    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, message: Assignation):
        try:
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            current_assignation.set(message)

            async for returns in self.assign(*args, **kwargs):

                returns = (
                    await shrink_outputs(
                        self.template.node,
                        returns,
                        structure_registry=self.structure_registry,
                    )
                    if self.shrink_outputs
                    else returns
                )

                await self.transport.change_assignation(
                    message.assignation, status=AssignationStatus.YIELD, results=returns
                )

            current_assignation.set(None)

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.DONE
            )

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCEL
            )

        except Exception as e:
            logger.error("Error in actor", exc_info=True)
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=repr(e)
            )

            raise e


class FunctionalThreadedFuncActor(FunctionalActor):
    def __init__(self, provision, *args, nworkers=4, **kwargs) -> None:
        super().__init__(provision, *args, **kwargs)
        self.threadpool = ThreadPoolExecutor(
            nworkers, thread_name_prefix=f"actor-thread-{provision.provision}"
        )

    async def on_assign(self, message: Assignation):

        try:
            logger.info("Assigning Number two")
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.ASSIGNED,
            )
            print("RUnnindosindfosidnfoisndfoin")
            returns = await run_spawned(self.assign, *args, **kwargs, pass_context=True)
            print("oinsoisnosinsoinsoin")
            shrinked_returns = (
                await shrink_outputs(
                    self.template.node,
                    returns,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.RETURNED,
                returns=shrinked_returns,
            )

        except asyncio.CancelledError as e:
            print("Canelled here")

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCELLED, message=str(e)
            )

        except Exception as e:
            print("Actor cancelled with exception", repr(e))
            logger.exception(e)
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=str(e)
            )

        print("Done in assignation")


class FunctionalThreadedGenActor(FunctionalActor):
    def __init__(self, provision, *args, nworkers=4, **kwargs) -> None:
        super().__init__(provision, *args, **kwargs)
        self.threadpool = ThreadPoolExecutor(
            nworkers, thread_name_prefix=f"actor-thread-{provision.provision}"
        )

    async def iterate_queue(
        self, async_q: janus._AsyncQueueProxy, message: Assignation
    ):
        try:
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]

                if action == "log":
                    raise NotImplementedError("Logging does not work right now")
                    async_q.task_done()
                if action == "yield":

                    returns = (
                        await shrink_outputs(
                            self.template.node,
                            value,
                            structure_registry=self.structure_registry,
                        )
                        if self.shrink_outputs
                        else value
                    )

                    await self.transport.change_assignation(
                        message.assignation,
                        status=AssignationStatus.YIELD,
                        returns=returns,
                    )
                    async_q.task_done()
                if action == "done":
                    await self.transport.change_assignation(
                        message.assignation,
                        status=AssignationStatus.DONE,
                    )
                    async_q.task_done()
                    break
                if action == "exception":
                    async_q.task_done()
                    raise value

        except asyncio.CancelledError as cancelled_error:
            while True:
                val = await async_q.get()
                action = val[0]
                value = val[1]
                if action == "exception":
                    async_q.task_done()
                    try:
                        raise value
                    except ThreadedActorCancelled:
                        raise cancelled_error

    def _assign_threaded(
        self,
        queue: janus._SyncQueueProxy,
        cancel_event_instance: threading.Event,
        message: Assignation,
        args: List[Any],
        kwargs: Dict[str, Any],
        parent_context: Optional[contextvars.Context],
    ):
        for var, value in parent_context.items():
            var.set(value)

        current_janus_queue.set(queue)
        current_assignation.set(message)
        current_cancel_event.set(cancel_event_instance)
        try:
            for result in self.assign(*args, **kwargs):
                queue.put(("yield", result))
                queue.join()

            queue.put(("done", "Happy doneness"))
            queue.join()

        except Exception as e:
            logger.exception(e)
            queue.put(("exception", e))
            queue.join()

        current_janus_queue.set(None)
        current_assignation.set(None)
        current_cancel_event.set(None)

    async def on_assign(self, message: Assignation):
        loop = asyncio.get_event_loop()
        queue = janus.Queue()
        event = threading.Event()

        try:
            logger.info("Assigning Number two")
            args, kwargs = (
                await expand_inputs(
                    self.template.node,
                    message.args,
                    message.kwargs,
                    structure_registry=self.structure_registry,
                )
                if self.expand_inputs
                else (message.args, message.kwargs)
            )

            await self.transport.change_assignation(
                message.assignation,
                status=AssignationStatus.ASSIGNED,
            )

            parent_context = contextvars.copy_context()

            threadedfut = loop.run_in_executor(
                self.threadpool,
                self._assign_threaded,
                queue.sync_q,
                event,
                message,
                args,
                kwargs,
                parent_context,
            )
            queuefut = self.iterate_queue(queue.async_q, message)

            try:
                await asyncio.gather(
                    asyncio.shield(threadedfut), asyncio.shield(queuefut)
                )
                queue.close()
                await queue.wait_closed()

            except asyncio.CancelledError as e:

                queuefut.cancel()  # We cancel the quefuture and are now only waiting for cancellation requests
                event.set()  # We are sending the request to the queue

                try:
                    await queuefut
                except asyncio.CancelledError as e:
                    raise e

        except asyncio.CancelledError as e:

            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CANCELLED, message=str(e)
            )

        except Exception as e:
            await self.transport.change_assignation(
                message.assignation, status=AssignationStatus.CRITICAL, message=str(e)
            )



from abc import abstractmethod
import asyncio
from typing import Dict, Optional


class Port:
    """
    Abstract class for serialization of data.

    """

    @abstractmethod
    async def cause_expand(self, value, registry):
        raise NotImplementedError()

    @abstractmethod
    async def cause_shrink(self, value, registry):
        raise NotImplementedError()


class ExpansionError:
    pass


class StructureExpander(Port):
    identifier: str

    async def cause_expand(self, value, registry):
        return await registry.get_expander_for_identifier(self.identifier)(value)

    async def cause_shrink(self, value, registry):
        return await registry.get_shrinker_for_identifier(self.identifier)(value)


class ListExpander(Port):
    child: Port

    async def cause_expand(self, value, registry):
        if self.child is None:
            raise ExpansionError("You need to specify a child")

        if isinstance(value, list):
            return await asyncio.gather(
                *[self.child.cause_expand(item, registry) for item in value]
            )
        else:
            raise ExpansionError(f"Expected a list got {type(value)}")

    async def cause_shrink(self, value, registry):
        return await asyncio.gather(
            *[self.child.cause_shrink(item, registry) for item in value]
        )


class IntExpander(Port):
    default_int: Optional[int]

    async def cause_expand(self, value, registry):
        if value == None:
            value = self.default_int
        if not isinstance(value, int):
            raise ExpansionError(f"Expected an int got {type(value)}")
        return value

    async def cause_shrink(self, value, registry):
        return int(value)


class StringExpander(Port):
    default_string: Optional[str]

    async def cause_expand(self, value, registry):
        if value == None:
            value = self.default_string
        if not isinstance(value, str):
            raise ExpansionError(f"Expected an str got {type(value)}")
        return value

    async def cause_shrink(self, value, registry):
        return str(value)


class DictExpander(Port):
    child: Port
    default_dict: Optional[Dict]

    async def cause_expand(self, value, registry):
        if value == None:
            value = self.default_dict
        if not isinstance(value, dict):
            raise ExpansionError(f"Expected an dict got {type(value)}")

        return {
            key: await self.child.cause_expand(value[key], registry) for key in value
        }

    async def cause_shrink(self, value, registry):
        return {
            key: await self.child.cause_shrink(value[key], registry) for key in value
        }


class EnumExpander(Port):
    default_enum: Optional[str]

    async def cause_expand(self, value, registry):
        if value == None:
            value = self.default_enum

        return value

    async def cause_shrink(self, value, registry):
        return str(value)


class BoolExpander(Port):
    default_bool: Optional[bool]

    async def cause_expand(self, value, registry):
        if value == None:
            value = self.default_bool

        return bool(value)

    async def cause_shrink(self, value, registry):
        return bool(value)



# JSON RPC Messages
import uuid
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional

from typing_extensions import Literal
from arkitekt.api.schema import AssignationStatus, ProvisionMode, ProvisionStatus
from arkitekt.messages import Assignation, Provision, Unassignation, Unprovision
from pydantic import BaseModel, Field


class AgentMessageTypes(str, Enum):

    ASSIGN_CHANGED = "ASSIGN_CHANGED"
    PROVIDE_CHANGED = "PROVIDE_CHANGED"

    LIST_ASSIGNATIONS = "LIST_ASSIGNATIONS"
    LIST_ASSIGNATIONS_REPLY = "LIST_ASSIGNATIONS_REPLY"
    LIST_ASSIGNATIONS_DENIED = "LIST_ASSIGNATIONS_DENIED"

    LIST_PROVISIONS = "LIST_PROVISIONS"
    LIST_PROVISIONS_REPLY = "LIST_PROVISIONS_REPLY"
    LIST_PROVISIONS_DENIED = "LIST_PROVISIONS_DENIED"


class AgentSubMessageTypes(str, Enum):

    ASSIGN = "ASSIGN"
    UNASSIGN = "UNASSIGN"
    PROVIDE = "PROVIDE"
    UNPROVIDE = "UNPROVIDE"


class JSONMeta(BaseModel):
    timestamp: datetime = Field(default_factory=datetime.utcnow)


class JSONMessage(BaseModel):
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    type: str
    meta: JSONMeta = Field(default_factory=JSONMeta)


class AssignationsList(JSONMessage):
    type: Literal[
        AgentMessageTypes.LIST_ASSIGNATIONS
    ] = AgentMessageTypes.LIST_ASSIGNATIONS
    exclude: Optional[List[AssignationStatus]]


class AssignationsListReply(JSONMessage):
    type: Literal[
        AgentMessageTypes.LIST_ASSIGNATIONS_REPLY
    ] = AgentMessageTypes.LIST_ASSIGNATIONS_REPLY
    assignations: List[Assignation]


class AssignationsListDenied(JSONMessage):
    type: Literal[
        AgentMessageTypes.LIST_ASSIGNATIONS_DENIED
    ] = AgentMessageTypes.LIST_ASSIGNATIONS_DENIED
    error: str


class ProvisionList(JSONMessage):
    type: Literal[AgentMessageTypes.LIST_PROVISIONS] = AgentMessageTypes.LIST_PROVISIONS
    exclude: Optional[List[AssignationStatus]]


class ProvisionListReply(JSONMessage):
    type: Literal[
        AgentMessageTypes.LIST_PROVISIONS_REPLY
    ] = AgentMessageTypes.LIST_PROVISIONS_REPLY
    provisions: List[Provision]


class ProvisionListDenied(JSONMessage):
    type: Literal[
        AgentMessageTypes.LIST_PROVISIONS_DENIED
    ] = AgentMessageTypes.LIST_PROVISIONS_DENIED
    error: str


class ProvisionChangedMessage(JSONMessage):
    type: Literal[AgentMessageTypes.PROVIDE_CHANGED] = AgentMessageTypes.PROVIDE_CHANGED
    provision: str
    status: Optional[ProvisionStatus]
    message: Optional[str]
    mode: Optional[ProvisionMode]


class AssignSubMessage(JSONMessage, Assignation):
    type: Literal[AgentSubMessageTypes.ASSIGN] = AgentSubMessageTypes.ASSIGN


class ProvideSubMessage(JSONMessage, Provision):
    type: Literal[AgentSubMessageTypes.PROVIDE] = AgentSubMessageTypes.PROVIDE


class UnassignSubMessage(JSONMessage, Unassignation):
    type: Literal[AgentSubMessageTypes.UNASSIGN] = AgentSubMessageTypes.UNASSIGN


class UnprovideSubMessage(JSONMessage, Unprovision):
    type: Literal[AgentSubMessageTypes.UNPROVIDE] = AgentSubMessageTypes.UNPROVIDE


class AssignationChangedMessage(JSONMessage):
    type: Literal[AgentMessageTypes.ASSIGN_CHANGED] = AgentMessageTypes.ASSIGN_CHANGED
    assignation: str
    status: Optional[AssignationStatus]
    message: Optional[str]
    returns: Optional[List[Any]]



import contextvars
from typing import Any
import janus

from arkitekt.messages import Assignation

current_assignation = contextvars.ContextVar("current_assignation")
transport = contextvars.ContextVar("transport")
current_janus_queue = contextvars.ContextVar("current_janus_queue")
current_provision_context = contextvars.ContextVar("current_provision_context")


def get_current_assignation() -> Assignation:
    return current_assignation.get()


def get_provision_context() -> Any:
    return current_provision_context.get()


def get_current_janus() -> janus.Queue:
    return current_janus_queue.get()



from typing import Any, Awaitable, Callable, Dict, List
from pydantic import Field
from arkitekt.api.schema import TemplateFragment, WidgetInput
from arkitekt.fakts.agent import FaktsAgent
from arkitekt.fakts.postman import FaktsPostman
from arkitekt.fakts.rath import FaktsArkitektRath
from arkitekt.rath import ArkitektRath
from arkitekt.messages import Provision
from arkitekt.structures.registry import (
    StructureRegistry,
    get_default_structure_registry,
)
from arkitekt.definition.registry import (
    DefinitionRegistry,
    get_current_definition_registry,
)
from arkitekt.agents.base import BaseAgent
from arkitekt.postmans.base import BasePostman
from koil import unkoil
from koil.composition import Composition
from koil.decorators import koilable


@koilable(fieldname="koil", add_connectors=True)
class Arkitekt(Composition):
    rath: ArkitektRath = Field(default_factory=FaktsArkitektRath)
    structure_registry: StructureRegistry = Field(
        default_factory=get_default_structure_registry
    )
    definition_registry: DefinitionRegistry = Field(
        default_factory=get_current_definition_registry
    )
    agent: BaseAgent = Field(default_factory=FaktsAgent)
    postman: BasePostman = Field(default_factory=FaktsPostman)

    def register(
        self,
        builder=None,
        widgets: Dict[str, WidgetInput] = {},
        interfaces: List[str] = [],
        on_provide: Callable[[Provision, TemplateFragment], Awaitable[Any]] = None,
        on_unprovide: Callable[[], Awaitable[Any]] = None,
        structure_registry: StructureRegistry = None,
        **actorparams,
    ) -> None:
        """
        Register a new function
        """
        structure_registry = structure_registry or self.structure_registry

        def real_decorator(function):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            self.definition_registry.register(
                function,
                builder=builder,
                widgets=widgets,
                interfaces=interfaces,
                structure_registry=structure_registry,
                on_provide=on_provide,
                on_unprovide=on_unprovide,
                **actorparams,
            )

        return real_decorator

    def run(self, *args, **kwargs) -> None:
        """
        Run the application.
        """
        return unkoil(self.arun, *args, **kwargs)

    async def arun(self) -> None:
        """
        Run the application.
        """
        assert self.agent.transport.connected, "Transport is not connected"
        await self.agent.aprovide()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        extra = "forbid"



import asyncio
from dataclasses import field
import json

import websockets
from arkitekt.postmans.transport.base import PostmanTransport
from typing import Any, Awaitable, Callable, Dict, List, Optional
from arkitekt.messages import Assignation, Reservation
from arkitekt.postmans.transport.errors import (
    AssignDeniedError,
    AssignationListDeniedError,
    PostmanTransportException,
    ReservationListDeniedError,
    ReserveDeniedError,
    UnassignDeniedError,
    UnreserveDeniedError,
)
from .protocols.postman_json import *
import logging
from websockets.exceptions import (
    ConnectionClosedError,
)

logger = logging.getLogger(__name__)


async def token_loader():
    raise NotImplementedError(
        "Websocket transport does need a defined token_loader on Connection"
    )


class CorrectableConnectionFail(PostmanTransportException):
    pass


class DefiniteConnectionFail(PostmanTransportException):
    pass


class WebsocketPostmanTransport(PostmanTransport):
    ws_url: str
    instance_id: Optional[str]
    token_loader: Callable[[], Awaitable[str]]
    max_retries = 5
    time_between_retries = 5
    allow_reconnect = True

    _futures: Dict[str, asyncio.Future] = field(default_factory=dict)
    _connected = False
    _healthy = False
    _send_queue: Optional[asyncio.Queue] = None
    _connection_task: Optional[asyncio.Task] = None

    async def __aenter__(self):
        assert (
            self.abroadcast is not None
        ), "Broadcast must be defined (either overwrite abroadcast or pass this in constructor of transport)"

        assert self.instance_id, "Needs an instance id"
        self._send_queue = asyncio.Queue()
        self._connection_task = asyncio.create_task(self.websocket_loop())
        self._connected = True

    async def websocket_loop(self, retry=0):
        send_task = None
        receive_task = None
        try:
            try:
                token = await self.token_loader()

                async with websockets.connect(
                    f"{self.ws_url}?token={token}&instance_id={self.instance_id}"
                ) as client:

                    logger.info("Postman on Websockets connected")

                    send_task = asyncio.create_task(self.sending(client))
                    receive_task = asyncio.create_task(self.receiving(client))

                    self._healthy = True
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self._healthy = True

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except ConnectionClosedError as e:
                logger.warning("Websocket was closed", exc_info=True)
                raise CorrectableConnectionFail from e

            except Exception as e:
                logger.warning("Websocket excepted. Trying to recover", exc_info=True)
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")
            if retry > self.max_retries or not self.allow_reconnect:
                raise DefiniteConnectionFail("Exceeded Number of Retries")

            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.websocket_loop(retry=retry + 1)

        except DefiniteConnectionFail as e:
            logger.error("Websocket excepted closed definetely", exc_info=True)
            self.connection_dead = False
            raise e

        except asyncio.CancelledError as e:
            logger.info("Websocket got cancelled. Trying to shutdown graceully")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            cancellation = await asyncio.gather(
                send_task, receive_task, return_exceptions=True
            )
            raise e

    async def sending(self, client):
        try:
            while True:
                message = await self._send_queue.get()
                await client.send(message)
                self._send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.info("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                await self.receive(message)
        except asyncio.CancelledError as e:
            logger.info("Receiving Task sucessfully Cancelled")

    async def receive(self, message):
        json_dict = json.loads(message)
        if "type" in json_dict:
            type = json_dict["type"]
            id = json_dict["id"]
            logger.debug(str(json_dict))

            # State Layer
            if type == PostmanSubMessageTypes.ASSIGN_UPDATE:
                await self.abroadcast(AssignSubUpdate(**json_dict))

            if type == PostmanSubMessageTypes.RESERVE_UPDATE:
                await self.abroadcast(ReserveSubUpdate(**json_dict))

            if type == PostmanMessageTypes.LIST_ASSIGNATION_REPLY:
                self._futures[id].set_result(AssingListReply(**json_dict))
            if type == PostmanMessageTypes.LIST_ASSIGNATION_DENIED:
                self._futures[id].set_exception(
                    AssignationListDeniedError(json_dict["error"])
                )

            if type == PostmanMessageTypes.LIST_RESERVATION_REPLY:
                self._futures[id].set_result(ReserveListReply(**json_dict))
            if type == PostmanMessageTypes.LIST_RESERVATION_DENIED:
                self._futures[id].set_exception(
                    ReservationListDeniedError(json_dict["error"])
                )

            if type == PostmanMessageTypes.ASSIGN_REPLY:
                self._futures[id].set_result(AssignPubReply(**json_dict))
            if type == PostmanMessageTypes.ASSIGN_DENIED:
                self._futures[id].set_exception(AssignDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.UNASSIGN_REPLY:
                self._futures[id].set_result(UnassignPubReply(**json_dict))
            if type == PostmanMessageTypes.UNASSIGN_DENIED:
                self._futures[id].set_exception(UnassignDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.RESERVE_REPLY:
                self._futures[id].set_result(ReservePubReply(**json_dict))
            if type == PostmanMessageTypes.RESERVE_DENIED:
                self._futures[id].set_exception(ReserveDeniedError(json_dict["error"]))

            if type == PostmanMessageTypes.UNRESERVE_REPLY:
                self._futures[id].set_result(UnreservePubReply(**json_dict))
            if type == PostmanMessageTypes.UNRESERVE_DENIED:
                self._futures[id].set_exception(
                    UnreserveDeniedError(json_dict["error"])
                )

        else:
            logger.error(f"Error {json_dict}")

    async def awaitaction(self, action: JSONMessage):
        assert self._connected, "Websocket not connected"
        if action.id in self._futures:
            raise ValueError("Action already has a future")

        future = asyncio.Future()
        self._futures[action.id] = future
        await self._send_queue.put(action.json())
        return await future

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        prov_list_reply: ReserveListReply = await self.awaitaction(
            ReserveList(exclude=exclude)
        )
        return prov_list_reply.reservations

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        ass_list_reply: AssignPubReply = await self.awaitaction(
            AssignPub(
                reservation=reservation,
                args=args,
                kwargs=kwargs,
                persist=persist,
                log=log,
            )
        )
        return ass_list_reply

    async def areserve(
        self, node: str, params: ReserveParamsInput = None
    ) -> Reservation:
        action = ReservePub(node=node, params=params)
        resrep: ReservePubReply = await self.awaitaction(action)
        return resrep

    async def aunreserve(self, reservation: str) -> Reservation:
        action = UnreservePub(reservation=reservation)
        unres: UnreservePubReply = await self.awaitaction(action)
        return unres

    async def aunassign(self, assignation: str) -> Reservation:
        action = UnassignPub(assignation=assignation)
        unass: UnassignPubReply = await self.awaitaction(action)
        return unass

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        action = AssignList(exclude=exclude)
        ass_list_reply: AssingListReply = await self.awaitaction(action)
        return ass_list_reply.assignations

    async def __aexit__(self, *args, **kwargs):
        self._connection_task.cancel()

        try:
            await self._connection_task
        except asyncio.CancelledError:
            pass



from typing import Any, Dict, List
from arkitekt.api.schema import NodeFragment
import asyncio
from arkitekt.structures.errors import ExpandingError, ShrinkingError
from arkitekt.structures.registry import StructureRegistry


async def shrink_inputs(
    node: NodeFragment,
    args: List[Any],
    kwargs: Dict[str, Any],
    structure_registry: StructureRegistry,
) -> List[Any]:
    """Shrinks args and kwargs

    Shrinks the inputs according to the Node Definition

    Args:
        node (Node): The Node

    Raises:
        ShrinkingError: If args are not Shrinkable
        ShrinkingError: If kwargs are not Shrinkable

    Returns:
        Tuple[List[Any], Dict[str, Any]]: Parsed Args as a List, Parsed Kwargs as a dict
    """
    if len(node.args) != len(args):
        raise ShrinkingError(f"Missmatch in Arg Length {node.args} vs {args}")

    shrinked_args_futures = [
        port.cause_shrink(arg, structure_registry) for port, arg in zip(node.args, args)
    ]

    try:
        shrinked_args = await asyncio.gather(
            *shrinked_args_futures
        )  # Worrysome because others won't be cancelled on Exception
    except Exception as e:
        raise ShrinkingError(
            f"Couldn't shrink Arguments {args} with {node.args}"
        ) from e

    try:
        shrinked_kwargs = {
            port.key: await port.cause_shrink(
                kwargs.get(port.key, None), structure_registry
            )
            for port in node.kwargs
        }
    except Exception as e:
        raise ShrinkingError(
            f"Couldn't shrink KeywordArguments {kwargs} with {node.kwargs}"
        ) from e

    return shrinked_args, shrinked_kwargs


async def expand_outputs(
    node: NodeFragment,
    returns: List[Any],
    structure_registry: StructureRegistry,
) -> List[Any]:
    """Expands Returns

    Expands the Returns according to the Node definition


    Args:
        node (Node): Node definition
        returns (List[any]): The returns

    Raises:
        ExpandingError: if they are not expandable

    Returns:
        List[Any]: The Expanded Returns
    """
    assert returns is not None, "Returns can't be empty"
    if len(node.returns) != len(returns):
        raise ExpandingError("Missmatch in Return Length")
    try:
        returns = await asyncio.gather(
            *[
                port.cause_expand(val, structure_registry)
                for port, val in zip(node.returns, returns)
            ]
        )

        if len(returns) == 1:
            return returns[0]  # We are dealing with a single output, just cast back
        else:
            return returns
    except Exception as e:
        raise ExpandingError(f"Couldn't expand Returns {returns}") from e



from typing import Callable, Dict, List, Optional, Tuple, Union

from pydantic import Field
from arkitekt.api.schema import TemplateFragment, acreate_template, adefine, afind
from arkitekt.definition.registry import (
    DefinitionRegistry,
    get_current_definition_registry,
)
from arkitekt.rath import ArkitektRath, current_arkitekt_rath
import asyncio
from arkitekt.agents.transport.base import AgentTransport
from arkitekt.messages import Assignation, Unassignation, Unprovision, Provision
from koil import unkoil
from koil.composition import KoiledModel


class BaseAgent(KoiledModel):
    """Agent

    Agents are the governing entities in the arkitekt system. They are responsible for
    spawning and managing the lifecycle of Actors and registering Templates with the help
    of the DefinitionRegistry.

    """

    transport: Optional[AgentTransport] = None
    definition_registry: Optional[DefinitionRegistry] = None
    rath: Optional[ArkitektRath] = None

    _approved_templates: List[Tuple[TemplateFragment, Callable]] = []
    _templateActorBuilderMap = {}
    _templateTemplatesMap: Dict[str, TemplateFragment] = {}
    _provisionActorMap = {}
    _provisionTaskMap: Dict[str, asyncio.Task] = Field(default_factory=dict)
    _inqueue: Optional[asyncio.Queue] = None

    async def abroadcast(
        self, message: Union[Assignation, Provision, Unassignation, Unprovision]
    ):
        await self._inqueue.put(message)

    async def process(self):
        raise NotImplementedError(
            "This method needs to be implemented by the agents subclass"
        )

    async def aregister_definitions(self):
        if self.definition_registry.templated_nodes:
            for (
                q_string,
                actor_builder,
                params,
            ) in self.definition_registry.templated_nodes:
                version = params.get("version", "main")

                arkitekt_node = await afind(q=q_string, rath=self.rath)

                arkitekt_template = await acreate_template(
                    node=arkitekt_node.id,
                    params=params,
                    version=version,
                    rath=self.rath,
                )

                self._approved_templates.append(
                    (arkitekt_template, actor_builder, params)
                )

        if self.definition_registry.defined_nodes:
            for (
                definition,
                actor_builder,
                params,
            ) in self.definition_registry.defined_nodes:
                # Defined Node are nodes that are not yet reflected on arkitekt (i.e they dont have an instance
                # id so we are trying to send them to arkitekt)
                arkitekt_node = await adefine(definition=definition, rath=self.rath)
                version = params.get("version", "main")
                arkitekt_template = await acreate_template(
                    node=arkitekt_node.id,
                    params={},  # Todo really make this happen
                    version=version,
                    rath=self.rath,
                )

                self._approved_templates.append(
                    (arkitekt_template, actor_builder, params)
                )

        if self._approved_templates:

            for arkitekt_template, defined_actor, params in self._approved_templates:
                # Generating Maps for Easy access
                self._templateActorBuilderMap[arkitekt_template.id] = defined_actor
                self._templateTemplatesMap[arkitekt_template.id] = arkitekt_template

    async def astep(self):
        await self.process(await self._inqueue.get())

    async def astart(self):
        await self.aregister_definitions()

        data = await self.transport.list_provisions()

        for prov in data:
            await self.abroadcast(prov)

        data = await self.transport.list_assignations()

        for ass in data:
            await self.abroadcast(ass)

    def step(self, *args, **kwargs):
        return unkoil(self.astep, *args, **kwargs)

    def start(self, *args, **kwargs):
        return unkoil(self.astart, *args, **kwargs)

    def provide(self, *args, **kwargs):
        return unkoil(self.aprovide, *args, **kwargs)

    async def aprovide(self):
        await self.astart()
        while True:
            await self.astep()

    async def __aenter__(self):
        self.definition_registry = (
            self.definition_registry or get_current_definition_registry()
        )
        self.rath = self.rath or current_arkitekt_rath.get()
        self._inqueue = asyncio.Queue()
        self.transport.abroadcast = self.abroadcast
        await self.transport.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.transport.__aexit__(exc_type, exc_val, exc_tb)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from pydantic import Field
from herre import Herre
from fakts import Fakts
from herre.fakts.herre import FaktsHerre
from koil.composition import Composition
from arkitekt.compositions.base import Arkitekt


class ArkitektApp(Composition):
    fakts: Fakts = Field(default_factory=Fakts)
    herre: Herre = Field(default_factory=FaktsHerre)
    arkitekt: Arkitekt = Field(default_factory=Arkitekt)



from typing import Optional

from arkitekt.postmans.transport.base import PostmanTransport
from arkitekt.postmans.vars import current_postman
from koil.composition import KoiledModel


class BasePostman(KoiledModel):
    transport: Optional[PostmanTransport] = None

    """Postman


    Postmans are the the messengers of the arkitekt platform, they are taking care
    of the communication between your app and the arkitekt-server.

    needs to implement:
        broadcast: On assignation Update or on reservation update (non updated fields are none)


    """

    async def __aenter__(self):
        current_postman.set(self)
        self.transport.abroadcast = self.abroadcast
        await self.transport.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.transport.__aexit__(exc_type, exc_val, exc_tb)
        current_postman.set(None)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arkitekt',
 'arkitekt.actors',
 'arkitekt.actors.testing',
 'arkitekt.agents',
 'arkitekt.agents.transport',
 'arkitekt.agents.transport.protocols',
 'arkitekt.api',
 'arkitekt.cli',
 'arkitekt.cli.dev',
 'arkitekt.cli.prod',
 'arkitekt.codegen',
 'arkitekt.compositions',
 'arkitekt.definition',
 'arkitekt.fakts',
 'arkitekt.postmans',
 'arkitekt.postmans.transport',
 'arkitekt.postmans.transport.protocols',
 'arkitekt.structures',
 'arkitekt.structures.serialization',
 'arkitekt.traits',
 'arkitekt.turms']

package_data = \
{'': ['*']}

install_requires = \
['docstring-parser>=0.11',
 'fakts==0.1.47',
 'herre>=0.1.69,<0.2.0',
 'inflection>=0.5.1,<0.6.0',
 'rath>=0.1.28,<0.2.0',
 'watchdog>=2.1.6,<3.0.0',
 'websockets>=10.0,<11.0']

entry_points = \
{'console_scripts': ['arkitekt = arkitekt.cli.main:entrypoint']}

setup_kwargs = {
    'name': 'arkitekt',
    'version': '0.1.119',
    'description': 'rpc and node backbone',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Awaitable, Callable, List, Optional, Dict, Any, Union
from arkitekt.messages import Assignation, Reservation, Unassignation, Unreservation
from arkitekt.api.schema import AssignationStatus, ReservationStatus, ReserveParamsInput
from koil.composition import KoiledModel


class PostmanTransport(KoiledModel):
    abroadcast: Optional[
        Callable[[Union[Assignation, Reservation]], Awaitable[None]]
    ] = None


    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        raise NotImplementedError()

    async def aunassign(self, assignation: str) -> Unassignation:
        raise NotImplementedError()

    async def areserve(
        self,
        node: str,
        params: ReserveParamsInput = None,
    ) -> Reservation:
        raise NotImplementedError()

    async def aunreserve(
        self,
        reservation: str,
    ) -> Unreservation:
        raise NotImplementedError()

    async def alist_assignations(
        self, exclude: Optional[AssignationStatus] = None
    ) -> List[Assignation]:
        raise NotImplementedError()

    async def alist_reservations(
        self, exclude: Optional[ReservationStatus] = None
    ) -> List[Reservation]:
        raise NotImplementedError()



from rich.table import Table


class Reserve:
    async def call_async_func(self, *args, reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            return await res.assign_async(*args, **kwargs)

    async def call_async_gen(self, *args, reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            async for result in res.stream_async(*args, **kwargs):
                yield result

    def __rich_repr__(self):
        yield self.name
        yield "args", self.args
        yield "kwargs", self.kwargs
        yield "returns", self.returns

    def __rich__(self):
        my_table = Table(title=f"Node: {self.name}", show_header=False)

        my_table.add_row("ID", str(self.id))
        my_table.add_row("Package", self.package)
        my_table.add_row("Interface", self.interface)
        my_table.add_row("Type", self.type)

        return my_table

    def _repr_html_(self):
        return f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.name}</div>
            <div class="item item-2">{self.package}/{self.interface}</div>
            <div class="item item-3">Args: {" ".join([port._repr_html_list() for port in self.args])}</div>
            <div class="item item-3">Kwargs: {" ".join([port.key for port in self.kwargs])}</div>
            <div class="item item-3">Returns: {" ".join([port.key for port in self.returns])}</div>
        </div>"""



from typing import Any, Dict, List, Union

from arkitekt.messages import Assignation, Reservation
from arkitekt.postmans.base import BasePostman
import asyncio
from pydantic import Field


class StatefulPostman(BasePostman):

    assignations: Dict[str, Assignation] = Field(default_factory=dict)
    reservations: Dict[str, Reservation] = Field(default_factory=dict)

    _res_update_queues: Dict[str, asyncio.Queue] = {}
    _ass_update_queues: Dict[str, asyncio.Queue] = {}

    async def __aenter__(self):
        await super().__aenter__()

        data = await self.transport.alist_reservations()
        self.reservations = {res.reservation: res for res in data}

        data = await self.transport.alist_assignations()
        self.assignations = {ass.assignation: ass for ass in data}

    async def areserve(self, node: str, params: dict = None) -> Reservation:
        reservation = await self.transport.areserve(node, params)
        self.reservations[reservation.reservation] = reservation
        return reservation

    async def aunreserve(self, reservation_id: str) -> Reservation:
        reservation = await self.transport.aunreserve(reservation_id)
        self.reservations[reservation.reservation] = reservation
        return reservation

    async def aassign(
        self,
        reservation: str,
        args: List[Any],
        kwargs: Dict[str, Any],
        persist=True,
        log=False,
    ) -> Assignation:
        assignation = await self.transport.aassign(
            reservation, args, kwargs, persist, log
        )
        self.assignations[assignation.assignation] = assignation
        return assignation

    def register_reservation_queue(self, res_id: str, queue: asyncio.Queue):
        self._res_update_queues[res_id] = queue

    def register_assignation_queue(self, ass_id: str, queue: asyncio.Queue):
        self._ass_update_queues[ass_id] = queue

    async def abroadcast(self, message: Union[Assignation, Reservation]):
        if isinstance(message, Assignation):
            self.assignations[message.assignation].update(message)
            if message.assignation in self._ass_update_queues:
                await self._ass_update_queues[message.assignation].put(
                    self.assignations[message.assignation]
                )
        elif isinstance(message, Reservation):
            self.reservations[message.reservation].update(message)
            if message.reservation in self._res_update_queues:
                await self._res_update_queues[message.reservation].put(
                    self.reservations[message.reservation]
                )

        else:
            raise Exception("Unknown message type")

    def unregister_reservation_queue(self, res_id: str):
        del self._res_update_queues[res_id]

    def unregister_assignation_queue(self, ass_id: str):
        del self._ass_update_queues[ass_id]



from typing import Any, Dict, List
from async_timeout import asyncio
from arkitekt.api.schema import NodeFragment
from arkitekt.structures.errors import ShrinkingError, ExpandingError
from arkitekt.structures.registry import StructureRegistry


async def expand_inputs(
    node: NodeFragment,
    args: List[Any],
    kwargs: Dict[str, Any],
    structure_registry: StructureRegistry,
):
    """Expand

    Args:
        node (NodeFragment): [description]
        args (List[Any]): [description]
        kwargs (List[Any]): [description]
        registry (Registry): [description]
    """

    expanded_args = []

    try:
        expanded_args = await asyncio.gather(
            *[
                port.cause_expand(arg, structure_registry)
                for port, arg in zip(node.args, args)
            ]
        )
    except Exception as e:
        raise ExpandingError(
            f"Couldn't expand Arguments {args} with {node.args}"
        ) from e

    try:
        expanded_kwargs = {
            port.key: await port.cause_expand(
                kwargs.get(port.key, None), structure_registry
            )
            for port in node.kwargs
        }
    except Exception as e:
        raise ExpandingError(
            f"Couldn't expand Kwargs {kwargs}  with {node.kwargs}"
        ) from e

    return expanded_args, expanded_kwargs


async def shrink_outputs(
    node: NodeFragment,
    returns: List[Any],
    structure_registry: StructureRegistry,
):
    """Expand

    Args:
        node (NodeFragment): [description]
        args (List[Any]): [description]
        kwargs (List[Any]): [description]
        registry (Registry): [description]
    """
    if returns is None:
        returns = ()
    if not isinstance(returns, tuple):
        returns = [returns]
    assert len(node.returns) == len(
        returns
    ), "Missmatch in Return Length"  # We are dealing with a single output, convert it to a proper port like structure

    shrinked_returns_future = [
        port.cause_shrink(val, structure_registry)
        for port, val in zip(node.returns, returns)
    ]
    try:
        return await asyncio.gather(*shrinked_returns_future)
    except Exception as e:
        raise ShrinkingError(f"Couldn't shrink Returns {returns}") from e



import asyncio
from importlib import import_module
from typing import List
import logging

logger = logging.getLogger(__name__)


async def retriebable_login(herre, max_retries=5, retry_period=5, retry=0):
    if retry == max_retries:
        raise Exception(f"We exceeded the retries for {herre}. Fatal Error!")
    try:
        await herre.alogin()
        print(f"{herre} Connected")
        return "Ok"
    except Exception as e:
        print(e)
        print(f"Sleeping for {retry_period}")
        await asyncio.sleep(retry_period)
        await retriebable_login(
            herre, max_retries=max_retries, retry_period=retry_period, retry=retry + 1
        )


async def wait_for_connection(modules: List[str]):
    """Wits for connection

    Import required Modules for this instance to run and tries to
    connect to the wards.

    TODO: The wards are not checked if alive at the same time
    at the end

    Args:
        modules (List[str]): The imported modules
    """

    ward_registry = get_ward_registry()

    imports = [import_module(module) for module in modules]

    herre = get_current_herre()
    await retriebable_login(herre)

    await asyncio.gather(*[retriable_connect(ward=ward) for ward in wardInstances])

    print("We are done whoop whoop")
